package com.hypr.server.commons.cloud.crypto;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.crypto.ECDSASigner;
import com.nimbusds.jose.jwk.Curve;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;

import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.time.Duration;
import java.util.Base64;
import java.util.Date;
import java.util.UUID;

public class JWTCreator {
    public static final String DEFAULT_JWT_AUDIENCE = "https://hypr.com";
    public static final long DEFAULT_JWT_MAX_VALIDITY_IN_SECS = 3600L;

    public static PrivateKey toPrivateKey(String privateKey) {
        try {
            return KeyFactory.getInstance("EC").generatePrivate(new PKCS8EncodedKeySpec(Base64.getDecoder().decode(privateKey)));
        } catch (InvalidKeySpecException | NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    public static String createSignedJwt(String subject, PrivateKey privateKey) {
        return createSignedJwt(subject, privateKey, DEFAULT_JWT_AUDIENCE, DEFAULT_JWT_MAX_VALIDITY_IN_SECS, null);
    }

    public static String createSignedJwt(String subject, String privateKey) {
        return createSignedJwt(subject, toPrivateKey(privateKey), DEFAULT_JWT_AUDIENCE, DEFAULT_JWT_MAX_VALIDITY_IN_SECS, null);
    }

    public static String createSignedJwt(String subject, PrivateKey privateKey, long validitySecs) {
        return createSignedJwt(subject, privateKey, DEFAULT_JWT_AUDIENCE, validitySecs, null);
    }

    public static String createSignedJwt(String subject, String privateKey, long validitySecs) {
        return createSignedJwt(subject, toPrivateKey(privateKey), DEFAULT_JWT_AUDIENCE, validitySecs, null);
    }

    public static String createSignedJwt(String subject, PrivateKey privateKey, long validitySecs, String keyId) {
        return createSignedJwt(subject, privateKey, DEFAULT_JWT_AUDIENCE, validitySecs, keyId);
    }

    public static String createSignedJwt(String subject, String privateKey, long validitySecs, String keyId) {
        return createSignedJwt(subject, toPrivateKey(privateKey), DEFAULT_JWT_AUDIENCE, validitySecs, keyId);
    }

    /**
     * @see [https://www.rfc-editor.org/rfc/rfc7523.html]
     */
    public static String createSignedJwt(String subject, PrivateKey privateKey, String audience, Long validForSecs, String keyId) {
        try {
            Date now = new Date();

            JWTClaimsSet claimsSet = new JWTClaimsSet.Builder()
                    // 3.1 The JWT MUST contain a "sub" (subject) claim identifying the principal that is the subject of the JWT
                    .subject(subject)

                    // 3.1 The JWT MUST contain an "iss" (issuer) claim that contains a unique identifier for the entity that issued the JWT
                    .issuer("hypr.com")

                    // 3.1 The JWT MUST contain an "aud" (audience) claim containing a value that identifies the authorization
                    // server as an intended audience
                    .audience(audience)

                    // 3.1 The JWT MAY contain a "jti" (JWT ID) claim that provides a unique identifier for the token
                    .jwtID(UUID.randomUUID().toString())

                    .issueTime(now)

                    // The JWT MAY contain an "nbf" (not before) claim that identifies the time before which the token MUST NOT be accepted for
                    // processing
                    .notBeforeTime(now)

                    // 3.1 The JWT MUST contain an "exp" (expiration time) claim that limits the time window during which the JWT can be used
                    .expirationTime(Date.from(now.toInstant().plus(Duration.ofSeconds(validForSecs))))

                    .build();

            final JWSHeader.Builder builder = new JWSHeader.Builder(JWSAlgorithm.ES256);
            if (keyId != null) {
                builder.keyID(keyId);
            }

            final SignedJWT signedJWT = new SignedJWT(builder.build(), claimsSet);

            // Create RSA-signer with the private key
            ECDSASigner signer = new ECDSASigner(privateKey, Curve.P_256);

            signedJWT.sign(signer);
            return signedJWT.serialize();
        } catch (JOSEException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * ℹ️
     * This file provides to JWT signing guidance to 3rd parties
     * Published on the HYPR doc portal
     */
    public static void main(String[] args) throws NoSuchAlgorithmException, InvalidKeySpecException {
        System.out.println("Usage ***************************************************************************\n" +
                           "java -cp \"*\" JWTCreation.java \"clientId\" \"clientPrivateKey\" \"keyId\" \n\n" +
                           "`clientId`   - required, get from token creation popup in the HYPR Control Center \n" +
                           "`privateKey` - required, to sign the JWT with. Base64 encoded EC or RSA key       \n" +
                           "`keyId`      - optional, to add a keyId to the JWT header. If you are using JWKS url, keyId is required. \n" +
                           "               Must match the keyId from the JWKS url response \n\n" +
                           "*********************************************************************************");

        //final String clientId = "hypap-fbc01c40-31c1-4fd9-b210-d075eadb1515";
        //final String privateKey = "MEECAQAwEwYHKoZIzj0CAQYIKoZIzj0DAQcEJzAlAgEBBCD6z+8VWzVyIHlECEEXW0zeWxAhDawPM983y4juGDEZBw==";
        //final String keyId = "8d266902-c637-4cf9-882a-076763822d3f";

        final String clientId = args[0];
        final String privateKey = args[1];
        final String keyId = args.length > 2 ? args[2] : null;

        final String assertion = JWTCreator.createSignedJwt(
                clientId,
                privateKey,
                3600,
                keyId);

        System.out.println("client_assertion:");
        System.out.println(assertion);
        System.out.println();
        System.out.println("You can get an OAuth API access token by exchanging this signed JWT");
        System.out.println();

        String formParams = String.format(
                "client_id=%s" + "&client_assertion=%s" + "&client_assertion_type=urn:ietf:params:oauth:client-assertion-type:jwt-bearer" +
                "&grant_type=client_credentials", clientId, assertion);

        System.out.printf("curl -X POST http://localhost:8009/rp/token/endpoint/exchange/clientcredentials"
                          + " -H \"Content-Type: application/x-www-form-urlencoded\""
                          + " -d \"%s\"", formParams);
    }
}
